/*
 * Copyright (C) 2008 Intel Corporation. All Rights Reserved.
 *
 * The source code contained or described herein and all
 * documents related to the source code ("Material") are owned by
 * Intel Corporation or its suppliers or licensors. Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors. The Material is protected by worldwide copyright
 * laws and treaty provisions.  No part of the Material may be
 * used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed,  or disclosed in any way
 * except as expressly provided in the license provided with the
 * Materials.  No license under any patent, copyright, trade
 * secret or other intellectual property right is granted to or
 * conferred upon you by disclosure or delivery of the Materials,
 * either expressly, by implication, inducement, estoppel or
 * otherwise, except as expressly provided in the license
 * provided with the Materials
*/
// [DESCRIPTION]
// 
// Solve the nqueens problem parallel exploration version
//
// [COMPILE]  
// Compile on IA-32 and Intel(R) 64 architectures only. 
// Use the following compiler options to compile: 
//
// CommandLine: icl /Qopenmp nq-parexp.cpp
//
//

#include <iostream>
#include <windows.h>

using namespace std;

// ------------------- Timing ----------------------

//****************************************************
//*  Windows NT (32 bit) dtime() routine             *
//*  Provided by: Piers Haken, piersh-at-microsoft.com  *
//****************************************************
#ifdef WIN32
#include <windows.h>

double dtime(void)
{
 double q;

 q = (double)GetTickCount() * 1.0e-03;

 return q;
}
#endif

//*****************************************************
//*  UNIX dtime(). This is the preferred UNIX timer.  *
//*****************************************************
#ifdef UNIX
#include <sys/time.h>
#include <sys/resource.h>

#ifdef hpux
#include <sys/syscall.h>
#define getrusage(a,b) syscall(SYS_getrusage,a,b)
#endif

struct rusage rusage;

double dtime()
{
 double q;

 getrusage(RUSAGE_SELF,&rusage);

 q = (double)(rusage.ru_utime.tv_sec);
 q = q + (double)(rusage.ru_utime.tv_usec) * 1.0e-06;

 return q;
}
#endif

// ------------------- Timing ----------------------

int nrOfSolutions=0;
int size=0;

void print(int queens[]) {
  cerr << "Solution " << nrOfSolutions << endl; 
  for(int row=0; row<size; row++) {
    for(int col=0; col<size; col++) {
      if(queens[row]==col) {
        cout << "Q";
      }
      else {
        cout << "-";
      }
    }
    cout << endl;
  }
}

void setQueen(int queens[], int row, int col) {
  for(int i=0; i<row; i++) {
    // vertical attacks
    if (queens[i]==col) {
      return;
    }
    // diagonal attacks
    if (abs(queens[i]-col) == (row-i) ) {
      return;
    }
  }

  // column is ok, set the queen
  queens[row]=col;

  if(row==size-1) {

#ifdef _DEBUG
    // only one thread should print allowed to print at a time
    __critical 
    {
      // increasing the solution counter is not atomic
      nrOfSolutions++;
      print(queens);
    }
#else 
    __critical 
    {
      // increasing the solution counter is not atomic(?)
      nrOfSolutions++;
    }
#endif
  }
  else {
    // try to fill next row
    for(int i=0; i<size; i++) {
      setQueen(queens, row+1, i);
    }
  }
}


void solve() {
    __par for(int i=0; i<size; i++) {
    // try all positions in first row
    // create separate array for each recursion
    // started here
    setQueen(new int[size], 0, i);
  }
}


int main(int argc, char*argv[]) {
  double startTime, endTime, dtime();
  if(argc !=2) {
    cerr << "Usage: nq-parexp boardSize.\n";
    return 0;
  }

  size = atoi(argv[1]);
  cout << "Starting Parallel Extension solver for size " << size << "...\n";

  startTime = dtime();
  solve();
  endTime = dtime() - startTime;
  endTime = endTime / 2.0;
  
  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took " << endTime << "ms.\n";

  return 0;
}